/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
#include <iostream>
#include <fstream>
#include <unordered_map>
#include "pin.H"

ofstream OutFile;

// The running count of instructions is kept here
// make it static to help the compiler optimize docount
static UINT64 icount = 0;
static UINT64 skipInstrCount = 0;
static UINT64 skipCount = 0;
static bool skipFlag;

UINT64 startIns;
UINT64 endIns;

static bool useHooks;
static bool inFastForward = true;

std::unordered_map<UINT64, UINT32> allocSize;
static UINT64 allocCount = 0;

#define ZSIM_MAGIC_OP_ROI_BEGIN (1025)
#define ZSIM_MAGIC_OP_ROI_END (1026)
#define ZSIM_MAGIC_OP_FFI_FF_EXIT (1050)
#define ZSIM_MAGIC_OP_SKIP_INSTR_BEGIN (2037)
#define ZSIM_MAGIC_OP_SKIP_INSTR_END (2038)
#define ZSIM_MAGIC_OP_RECORD_ALLOCATION (2039)

ADDRINT HandleMagicOp(ADDRINT op, CONTEXT* ctxt);
void dumpStats();

// This function is called before every instruction is executed
VOID docount() {
    if (useHooks) {
        if (!inFastForward) icount++;
    } else {
        icount++;
        if (icount >= endIns) {
            dumpStats();
            exit(0);
        }
    }
}

VOID skipInsCount() { skipInstrCount++; }

// Pin calls this function every time a new instruction is encountered
VOID Instruction(INS ins, VOID* v) {
    // Insert a call to docount before every instruction, no arguments are
    // passed
    if (!skipFlag) {
        INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)docount, IARG_END);
    } else {
        INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)skipInsCount,
                       IARG_FAST_ANALYSIS_CALL, IARG_END);
    }

    if (INS_IsXchg(ins) && INS_OperandReg(ins, 0) == REG_RCX &&
        INS_OperandReg(ins, 1) == REG_RCX) {
        INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)HandleMagicOp,
                       IARG_REG_VALUE, REG_ECX, IARG_CONTEXT, IARG_RETURN_REGS,
                       REG_EAX, IARG_END);
    }
}

ADDRINT HandleMagicOp(ADDRINT op, CONTEXT* ctxt) {
    switch (op) {
        case ZSIM_MAGIC_OP_ROI_BEGIN:
            if (useHooks) {
                inFastForward = false;
            }
            return 0;

        case ZSIM_MAGIC_OP_ROI_END:
            if (useHooks) {
                inFastForward = true;
                dumpStats();
                exit(0);
            }

        /*
        case ZSIM_MAGIC_OP_FFI_FF_EXIT:
            inFastForward = false
        */

        case ZSIM_MAGIC_OP_SKIP_INSTR_BEGIN:
            if (!skipFlag) {
                skipFlag = 1;
                skipCount++;
            }
            break;

        case ZSIM_MAGIC_OP_SKIP_INSTR_END:
            if (skipFlag) {
                skipFlag = 0;
            }
            break;

        case ZSIM_MAGIC_OP_RECORD_ALLOCATION:
            UINT64 hash = PIN_GetContextReg(ctxt, REG_RSI);
            UINT32 size = PIN_GetContextReg(ctxt, REG_RDX);
            if (allocSize.count(hash))
                allocSize[hash] = allocSize[hash] + size;
            else
                allocSize[hash] = size;
            allocCount++;
            break;
    }
    return 0;
}

KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE, "pintool", "o", "inscount.out",
                            "specify output file name");

KNOB<uint64_t> KnobROIStart(KNOB_MODE_WRITEONCE, "pintool", "start",
                                 "20000000000", "start of ROI(instructions)");

KNOB<uint64_t> KnobROILength(KNOB_MODE_WRITEONCE, "pintool", "length",
                                  "10000000000", "Length of ROI(instructions)");

KNOB<bool> KnobHooks(KNOB_MODE_WRITEONCE, "pintool", "hooks", "0",
                     "use roi hooks in application instead of instructions based sampling");

void dumpStats() {
    // Write to a file since cout and cerr maybe closed by the application
    OutFile << allocCount << endl;
    for (auto it = allocSize.begin(); it != allocSize.end(); it++) {
        OutFile << std::hex << it->first << " " << std::dec << it->second
                << endl;
    }
    OutFile.close();

    return;
}

// This function is called when the application exits
VOID Fini(INT32 code, VOID* v) { dumpStats(); }

/* ===================================================================== */
/* Print Help Message                                                    */
/* ===================================================================== */

INT32 Usage() {
    cerr << "This tool is used with poolmalloc library to instrument "
            "allocations";
    cerr << " in a specific region of interest(ROI) specified in terms of "
            "instructions" << endl;
    cerr << endl << KNOB_BASE::StringKnobSummary() << endl;
    return -1;
}

/* ===================================================================== */
/* Main                                                                  */
/* ===================================================================== */
/*   argc, argv are the entire command line: pin -t <toolname> -- ...    */
/* ===================================================================== */

int main(int argc, char* argv[]) {
    // Initialize pin
    if (PIN_Init(argc, argv)) return Usage();

    OutFile.open(KnobOutputFile.Value().c_str());

    startIns = KnobROIStart.Value();
    endIns = startIns + KnobROILength.Value();
    useHooks = KnobHooks.Value();

    // Register Instruction to be called to instrument instructions
    INS_AddInstrumentFunction(Instruction, 0);

    // Register Fini to be called when the application exits
    PIN_AddFiniFunction(Fini, 0);

    // Start the program, never returns
    PIN_StartProgram();

    return 0;
}
