/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/

#define UNW_LOCAL_ONLY
#include "poolMalloc.h"
#include "libc_override.h"
#include "simpleHash.h"

#include <assert.h>
#include <cstring>
#include <dlfcn.h>
#include <execinfo.h>
#include <functional>
#include <fstream>
#include <libunwind.h>
#include <pthread.h>
#include <signal.h>
#include <syscall.h>
#include <unistd.h>
#include <unordered_map>
#include <utility>
#include <vector>

#define Ptr(pool) pmstate.poolPtrs[pool]
#define ensure_initialization() (void)(pmstate.initialized != 0 || do_initialization())

#define likely(x)  __builtin_expect((x), 1)
#define unlikely(x) __builtin_expect((x), 0)

// Custom allocator class that calls pool_malloc variants
// This allocator is used by the unordered_map that maps callpoints to pools
// Source : https://github.com/s5z/zsim/blob/master/src/g_std/stl_galloc.h
template <typename T>
class StlAlloc {
    public:
        typedef size_t size_type;
        typedef ptrdiff_t difference_type;
        typedef T* pointer;
        typedef const T* const_pointer;
        typedef T& reference;
        typedef const T& const_reference;
        typedef T value_type;

        StlAlloc() {
        }
        StlAlloc(const StlAlloc&) {
        }
        pointer allocate(size_type n, const void * = 0) {
            T* t =  static_cast<T*>(pool_malloc(sizeof(T)*n));
            return t;
        }

        void deallocate(void* p, size_type) {
            if (p) pool_free(p);
        }

        pointer address(reference x) const { return &x; }
        const_pointer address(const_reference x) const { return &x; }
        StlAlloc<T>& operator=(const StlAlloc&) { return *this; }


        // Construct/destroy
        // gcc keeps changing these interfaces. See /usr/include/c++/4.8/ext/new_allocator.h
#if __cplusplus >= 201103L // >= 4.8
        template<typename _Up, typename... _Args>
        void construct(_Up* __p, _Args&&... __args) { ::new((void *)__p) _Up(std::forward<_Args>(__args)...); }

        template<typename _Up> void destroy(_Up* __p) { __p->~_Up(); }
#else // < 4.8
        void construct(pointer p, const T& val) { new (static_cast<T*>(p)) T(val); }
        void construct(pointer p) { construct(p, value_type()); } //required by gcc 4.6
        void destroy(pointer p) { p->~T(); }
#endif

        size_type max_size() const { return size_t(-1); }

        template <class U> struct rebind { typedef StlAlloc<U> other; };

        template <class U> StlAlloc(const StlAlloc<U>&) {}

        template <class U> StlAlloc& operator=(const StlAlloc<U>&) { return *this; }

        template <class U> bool operator==(const StlAlloc<U>&) const { return true; }

        template <class U> bool operator!=(const StlAlloc<U>&) const { return false; }
};

// This contains all the relevant state of the runtime
// like knobs, callpoint->pool maps etc.
struct poolMallocState {
    uint32_t numPools; // Excluding thread pool
    uint32_t numPCs;
    uint32_t maxPools;
    void** poolPtrs; // pointers to mspaces

    bool initialized;
    bool allocPoolSwitch; // flag to disambiguate allocation calls from within poolMalloc
                          // routines(false) and those from application-level code(true).

    bool mode; // 0=profile, 1=exec
    bool signature; // how to generate callpoint ids? 0=hash of stacktrace, 1=calling pc
    size_t traceDepth; // depth of stacktrace when signature = 0
    bool relAddr; // 1 = disambiguate addresses from shared libraries

    // State used with libunwind routines
    Dl_info* dlinfo;
    unw_cursor_t cursor;
    unw_context_t uc;
    unw_word_t ip;

    void* array[6]; // uint64_t for normal stepping, void* for backtrace function
    size_t bsize;

    // Exec mode state
    std::unordered_map<uint64_t, uint32_t, std::hash<uint64_t>, std::equal_to<uint64_t>, StlAlloc< std::pair<const uint64_t, uint32_t> >>* hashToPoolMap;

    // Profile mode state
    std::unordered_map<uint64_t, uint32_t, std::hash<uint64_t>, std::equal_to<uint64_t>, StlAlloc< std::pair<const uint64_t, uint32_t> >>* hashList;
    uint64_t totalAllocCount;

    // thread-safety
    pthread_mutex_t lock;
    int64_t curTid; // thread_id of the thread that holds lock.
                    // this is to ensure that each thread only tries to grab the lock
                    // once per call to pool_malloc/pool_free routines
};


/* Global Variables */
static struct poolMallocState pmstate;

/* Forward declarations */
static int loadPCTable();
void got_signal(int);

static int do_initialization() {
    if (pthread_mutex_init(&pmstate.lock, NULL) != 0){
        printf("mutex init failed\n");
    }

    int myTid = syscall(SYS_gettid);
    pthread_mutex_lock(&pmstate.lock);
    pmstate.curTid = myTid;

    pmstate.numPools = 1;
    pmstate.numPCs = 0;
    pmstate.maxPools = 512;
    void* ptr = create_mspace(0,1, false);

    pmstate.poolPtrs = (void**) mspace_malloc(ptr, sizeof(void*)*(pmstate.maxPools+1));
    pmstate.poolPtrs[0] = ptr;

    pmstate.initialized = true; // Bare minimum initialization done

    struct sigaction sa;
    memset( &sa, 0, sizeof(sa) );
    sa.sa_handler = got_signal;
    sigfillset(&sa.sa_mask);
    sigaction(SIGINT,&sa,NULL);

    char* mode = getenv("PM_MODE");
    if (mode != NULL) pmstate.mode = (mode[0] != '0');
    char* signature = getenv("PM_SIGNATURE");
    if (signature != NULL) pmstate.signature = (signature[0] != '0');
    char* traceDepth = getenv("PM_TRACEDEPTH");
    if (traceDepth != NULL) pmstate.traceDepth = atoi(traceDepth);
    if (pmstate.traceDepth == 0) pmstate.traceDepth = 3; // default value
    char* relAddr = getenv("PM_RELADDR");
    if (relAddr != NULL) pmstate.relAddr= atoi(relAddr);
    else if (!pmstate.signature) pmstate.relAddr = 1; // by default take relative addresses for stack traces

    fprintf(stderr, "[poolMalloc] mode is %d\n", pmstate.mode);
    fprintf(stderr, "[poolMalloc] signature is %d\n", pmstate.signature);
    fprintf(stderr, "[poolMalloc] traceDepth is %lu\n", pmstate.traceDepth);
    fprintf(stderr, "[poolMalloc] relAddr is %d\n", pmstate.relAddr);

    pmstate.dlinfo = (Dl_info*) mspace_malloc(ptr, sizeof(Dl_info));

    pmstate.hashToPoolMap = new std::unordered_map<uint64_t, uint32_t, std::hash<uint64_t>, std::equal_to<uint64_t>, StlAlloc< std::pair<const uint64_t, uint32_t> > >;
    pmstate.hashList = new std::unordered_map<uint64_t, uint32_t, std::hash<uint64_t>, std::equal_to<uint64_t>, StlAlloc< std::pair<const uint64_t, uint32_t> > >;

    if (pmstate.mode) {
        if (loadPCTable()) { fprintf(stderr, "[poolMalloc] Initialization failed\n"); exit(0); }
    }

    fprintf(stderr, "[poolMalloc] Initialization done\n");

    pmstate.allocPoolSwitch = true; // Complete Intialization done
    pthread_mutex_unlock(&pmstate.lock);

    return 1;
}

static int loadPCTable() {
    char* pcTableFile = getenv("PM_MAPPINGSFILE");
    fprintf(stderr, "[poolMalloc] pcTableFile %s\n", pcTableFile);

    if (pcTableFile == NULL) {
        fprintf(stderr, "[poolMalloc] Empty pcTableFile\n");
        return 1;
    }

    FILE* fp = fopen(pcTableFile, "r");
    if (fp == NULL) { fprintf(stderr, "[poolMalloc] Error opening pcTable file\n"); return 1; }

    uint32_t numPools;
    if (!(fscanf(fp, " %u", &numPools))) { fprintf(stderr, "[poolMalloc] Bad pcTable format numPools\n"); return 1; }

    if (numPools == 0) return 0;
    if (numPools > pmstate.maxPools) { fprintf(stderr, "[poolMalloc] numPools greater than maxPools\n"); return 1; }

    uint32_t numPCs;
    if (!(fscanf(fp, " %u", &numPCs))) { fprintf(stderr, "[poolMalloc] Bad pcTable format numPCs\n"); return 1; }

    int line = 1;
    int i = 0;
    while (!feof(fp)) {
        uint64_t curPC;
        uint32_t curPool;
        int count = fscanf(fp, " %lx %u", &curPC, &curPool);
        if (curPool > numPools) {
            fprintf(stderr, "[poolMalloc] Invalid pool line no %d\n", line); return 1;
        }
        if (count != 2) { fprintf(stderr, "[poolMalloc] Bad pcTable format line no %d\n", line); return 1; }
        pmstate.hashToPoolMap->insert({curPC, curPool});
        line++;
        i++;
        fscanf(fp, " ");
    }
    pmstate.numPools = numPools;
    pmstate.numPCs = numPCs;

    // Create memory pools
    for (int i = 1; i <= pmstate.numPools; i++) {
        void* ptr = create_mspace(0, 1, true);
        pmstate.poolPtrs[i] = ptr;
    }

    return 0;
}

// This is called by all wrapper functions to compute the pool/mspace from which to make allocations.
__attribute__((always_inline)) inline uint32_t get_alloc_pool(uint32_t size)  {
    zsim_skip_instr_begin();
    uint64_t hashValue = 0;

    if (!pmstate.signature) { // Stack traces
        pmstate.bsize = unw_backtrace(pmstate.array, 1 + pmstate.traceDepth);
        // Get relative adddresses
        if (pmstate.relAddr) {
            for (int i = 1; i < pmstate.bsize; i++) {
                if(dladdr((void*)pmstate.array[i], pmstate.dlinfo)) {
                    pmstate.array[i] = (void*)((uint64_t)pmstate.array[i] - (uint64_t)pmstate.dlinfo->dli_fbase);
                }
            }
        }

        // Ignore one-level to account for backtrace callpoint address(point where we call backtrace above)
        hashValue = hashword((uint64_t*)pmstate.array + 1, (pmstate.bsize >= pmstate.traceDepth + 1) ? pmstate.traceDepth : pmstate.bsize - 1, 0);

    } else { // PC-based
        hashValue = (uint64_t) __builtin_return_address(0); // 0 because we inlined get_alloc_pool()
        // Get relative adddress
        if (pmstate.relAddr) {
            if(dladdr((void*)hashValue, pmstate.dlinfo)){
                uint64_t base = ((uint64_t)pmstate.dlinfo->dli_fbase == 0x400000) ? 0 : (uint64_t)pmstate.dlinfo->dli_fbase;
                hashValue = hashValue - base;
            }
        }
    }

    if (likely(pmstate.mode)) {
        std::unordered_map<uint64_t, uint32_t, std::hash<uint64_t>, std::equal_to<uint64_t>, StlAlloc< std::pair<const uint64_t, uint32_t> >>::iterator it = pmstate.hashToPoolMap->find(hashValue);
        if (it != pmstate.hashToPoolMap->end()) { zsim_skip_instr_end(); return it->second; }
    } else {
        if (size > 0) {
            pmstate.totalAllocCount++;
            if (!pmstate.hashList->count(hashValue)) pmstate.hashList->insert({hashValue, size});
            else pmstate.hashList->at(hashValue) += size;
            zsim_record_allocation(hashValue, size);
         }
    }
    zsim_skip_instr_end();
    return 0; // default allocations go to thread-pool
}

// MACROS related to thread safety
#define get_lock(); \
    int myTid = syscall(SYS_gettid); \
    bool localMutexFlag = 0; \
    if (pmstate.curTid == 0) { \
        pthread_mutex_lock(&pmstate.lock); \
        pmstate.curTid = myTid; \
        localMutexFlag = 1; \
    } else if (pmstate.curTid != myTid) { \
        pthread_mutex_lock(&pmstate.lock); \
        pmstate.curTid = myTid; \
        localMutexFlag = 1;\
    } \

#define release_lock(); \
    if (localMutexFlag) { \
        pmstate.curTid = 0; \
        pthread_mutex_unlock(&pmstate.lock); \
    } \

// Uncomment these if you don't need thread safety.
//#define get_lock();
//#define release_lock();

void* pool_malloc(size_t size) __THROW {
    ensure_initialization();
    get_lock();

    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_malloc(pmstate.poolPtrs[get_alloc_pool(size)], size);
        release_lock();
        pmstate.allocPoolSwitch = true;
        return ptr;
    } else {
        void* ptr = mspace_malloc(pmstate.poolPtrs[0], size);
        release_lock();
        return ptr;
   }

}

void pool_free(void* ptr) __THROW {
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        dlfree(ptr); // FOOTERS = 1
        pmstate.allocPoolSwitch = true;
        release_lock();
    } else {
        dlfree(ptr); // FOOTERS = 1
        release_lock();
    }
}

void pool_cfree(void* ptr) __THROW {
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        dlfree(ptr); // FOOTERS = 1
        pmstate.allocPoolSwitch = true;
        release_lock();
    } else {
        dlfree(ptr); // FOOTERS = 1
        release_lock();
    }
}

void* pool_calloc(size_t num, size_t size) __THROW {
    ensure_initialization();
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_calloc(pmstate.poolPtrs[get_alloc_pool(num*size)], num, size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_calloc(pmstate.poolPtrs[0], num, size);
        release_lock();
        return ptr;
    }
}

void* pool_realloc(void* ptr, size_t newSz) __THROW {
    ensure_initialization();
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* newPtr = dlrealloc(ptr, newSz); // FOOTERS = 1
        pmstate.allocPoolSwitch = true;
        release_lock();
        return newPtr;
    } else {
        void* newPtr = dlrealloc(ptr, newSz); // FOOTERS = 1
        release_lock();
        return newPtr;
    }
}

int pool_posix_memalign(void **memptr, size_t alignment , size_t size) __THROW {
    ensure_initialization();
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        *memptr = mspace_memalign(pmstate.poolPtrs[get_alloc_pool(size)], alignment, size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        if (!*memptr) return 12; // ENOMEM
        return 0;
    } else {
        *memptr = mspace_memalign(pmstate.poolPtrs[0], alignment, size);
        release_lock();
        if (!*memptr) return 12; // ENOMEM
        return 0;
    }

}

void* pool_valloc(size_t size) __THROW {
    ensure_initialization();
    get_lock();

    size_t page_size;
    #ifdef _SC_PAGE_SIZE
        page_size = _SC_PAGE_SIZE;
    #else
        page_size = sysconf(_SC_PAGE_SIZE);
    #endif

    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_memalign(pmstate.poolPtrs[get_alloc_pool(size)], page_size, size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_memalign(pmstate.poolPtrs[0], page_size, size);
        return ptr;
        release_lock();
    }
}

void* pool_pvalloc(size_t size) __THROW {
    ensure_initialization();
    get_lock();

    size_t page_size;
    #ifdef _SC_PAGE_SIZE
        page_size = _SC_PAGE_SIZE;
    #else
        page_size = sysconf(_SC_PAGE_SIZE);
    #endif
     if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_memalign(pmstate.poolPtrs[get_alloc_pool(size)], page_size, (size + page_size - (size_t) 1) & ~(page_size - (size_t)1) );
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
     } else {
        void* ptr = mspace_memalign(pmstate.poolPtrs[0], page_size, (size + page_size - (size_t) 1) & ~(page_size - (size_t)1) );
        release_lock();
        return ptr;
     }
}

void* pool_memalign(size_t alignment, size_t size) __THROW {
    ensure_initialization();
    get_lock();

    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_memalign(pmstate.poolPtrs[get_alloc_pool(size)], alignment, size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_memalign(pmstate.poolPtrs[0], alignment, size);
        release_lock();
        return ptr;
    }
}

void* pool_new(size_t size) {
    ensure_initialization();
    get_lock();

    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_malloc(pmstate.poolPtrs[get_alloc_pool(size)], size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_malloc(pmstate.poolPtrs[0], size);
        release_lock();
        return ptr;
    }
}

void pool_delete(void* ptr) __THROW {
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        dlfree(ptr);
        pmstate.allocPoolSwitch = true;
        release_lock();
    } else {
        dlfree(ptr);
        release_lock();
    }
}

void* pool_newarray(size_t size) {
    ensure_initialization();
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_malloc(pmstate.poolPtrs[get_alloc_pool(size)], size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_malloc(pmstate.poolPtrs[0], size);
        release_lock();
        return ptr;
    }
}

void pool_deletearray(void* ptr) __THROW {
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        dlfree(ptr);
        pmstate.allocPoolSwitch = true;
        release_lock();
    } else {
        dlfree(ptr);
        release_lock();
    }
}

void* pool_new_nothrow(size_t size, const std::nothrow_t& nt) __THROW {
    ensure_initialization();
    get_lock();

    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_malloc(pmstate.poolPtrs[get_alloc_pool(size)], size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_malloc(pmstate.poolPtrs[0], size);
        release_lock();
        return ptr;
    }
}

void* pool_newarray_nothrow(size_t size, const std::nothrow_t& nt) __THROW {
    ensure_initialization();
    get_lock();

    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        void* ptr = mspace_malloc(pmstate.poolPtrs[get_alloc_pool(size)], size);
        pmstate.allocPoolSwitch = true;
        release_lock();
        return ptr;
    } else {
        void* ptr = mspace_malloc(pmstate.poolPtrs[0], size);
        release_lock();
        return ptr;
    }
}

void pool_delete_nothrow(void* ptr, const std::nothrow_t& nt) __THROW {
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        dlfree(ptr);
        pmstate.allocPoolSwitch = true;
        release_lock();
    } else {
        dlfree(ptr);
        release_lock();
    }
}

void pool_deletearray_nothrow(void* ptr, const std::nothrow_t& nt) __THROW {
    get_lock();
    if (pmstate.allocPoolSwitch) {
        pmstate.allocPoolSwitch = false;
        dlfree(ptr);
        pmstate.allocPoolSwitch = true;
        release_lock();
    } else {
        dlfree(ptr);
        release_lock();
    }
}

// Dumps the callpoint -> total bytes allocated information
__attribute__((destructor)) void dumpHashList() {
    pmstate.allocPoolSwitch = false;
    if (!pmstate.mode) {
        char* outputFile = getenv("PM_OUTPUTFILE");
        fprintf(stderr, "[poolMalloc] output file is: %s \n", outputFile);
        FILE* fp = fopen(outputFile, "w");
        if (fp == NULL) { fprintf(stderr, "[poolMalloc] Error opening output file \n"); exit(0); }
        fprintf(fp, "%lu\n", pmstate.totalAllocCount);
        fprintf(fp, "%lu\n", pmstate.hashList->size());
        for (auto it = pmstate.hashList->begin(); it != pmstate.hashList->end(); it++){
            fprintf(fp, "%lx %u\n", it->first, it->second);
        }
        fclose(fp);
    }
    return;
}

// dump hashList before dying
void got_signal(int){
    dumpHashList();
    fprintf(stderr, "[poolMalloc] Hash List dumped on SIGINT \n");
    exit(1);
}
