#!/usr/bin/python
'''
/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
'''

import h5py
import numpy as np
from scipy.cluster.hierarchy import fcluster

from baseScript import args, tableau_colors, InitGraphsDir, sizeInMB
from whirltoolHelper import printClusters, readPcTableFile, optimizePools,\
    genPcTableFile, writePcTableFile, writeClusterDistances
from hcluster import wpLinkage, cluster_node

### Read input files and setup output directories ###

dset = h5py.File(args.inputFile, 'r')

phases = dset['shareMissCurves'].shape[0]
numShares = dset['shareMissCurves'].shape[1]
sysMissCurvePoints = dset['shareMissCurves'].shape[2]

### Read miss curve, sample the time and capacity axes if needed and get MPKI values ###

startPhase = 1
endPhase = phases

shareMissCurves = dset['shareMissCurves'][startPhase:endPhase].astype(float)
shareMissCurvePoints = shareMissCurves.shape[2]

## Scale capacity axis to have a fixed resolution per MB of cache size
llcSize = sizeInMB(args.llcSize)
bucketsPerMB = int((shareMissCurvePoints-1)/llcSize)
if bucketsPerMB > 16: # lets use 64KB buckets
    capacityAxisSamplingFactor = int(bucketsPerMB/16.0)
else:
    capacityAxisSamplingFactor = 1

print "Scaling capacity axis by %d : from %d to %d"\
       % (capacityAxisSamplingFactor, shareMissCurvePoints, (shareMissCurvePoints-1)/capacityAxisSamplingFactor + 1)

shareMissCurves = shareMissCurves[:,:,::capacityAxisSamplingFactor]
shareMissCurvePoints = shareMissCurves.shape[2]

## Scale time axis to have a maximum of 1024 time samples
totalTimeSamples = shareMissCurves.shape[0]
if totalTimeSamples > 1024:
    timeAxisSamplingFactor = totalTimeSamples/1024 + 1
    shareMissCurves = shareMissCurves[::timeAxisSamplingFactor, :, :]
    finalTimeSamples = shareMissCurves.shape[0]
    print "Scaling time axis by %d : from %d to %d"\
        % (timeAxisSamplingFactor, totalTimeSamples, finalTimeSamples)
    assert(finalTimeSamples > 1024/2 and finalTimeSamples <= 1024)

epochInstrs = 50000000
## Get MPKI values
for time in range(shareMissCurves.shape[0]):
    shareMissCurves[time] /= epochInstrs

### Start clustering ###

## Discard pools with very-low apki
cumAccesses = np.sum(shareMissCurves[:, :, 0], axis = 0)
cumAccesses /= np.sum(cumAccesses)

# tune args.threshold to get a maximum of 50 pools
print "Low apki threshold tuning"
highAPKIpools = len(cumAccesses)
while True:
    lowAPKIpools = cumAccesses < args.threshold
    highAPKIpools = len(lowAPKIpools) - sum(lowAPKIpools)
    discardedPoolApki = sum([cumAccesses[i] for i in range(numShares) if lowAPKIpools[i]])
    print highAPKIpools, discardedPoolApki

    if discardedPoolApki > 0.05 or highAPKIpools <= 50:
        if discardedPoolApki > 0.1:
            args.threshold /= 2
            lowAPKIpools = cumAccesses < args.threshold
        break
    else:
        args.threshold *= 2

print "Final low apki threshold", args.threshold

highAPKIindices = [i for i in range(numShares) if not lowAPKIpools[i]]
print "Discarding %d low APKI pools" % (numShares - len(highAPKIindices))
print "Considering pools", highAPKIindices
sortedByAPKI = np.argsort(cumAccesses[highAPKIindices])[::-1]
print "Pools sorted by APKI", [highAPKIindices[i] for i in sortedByAPKI]
print "Cumulative accesses of high APKI pools", [cumAccesses[highAPKIindices[i]] for i in sortedByAPKI]

pcList, poolList = readPcTableFile(args.mappingsFile)
print "Callpoints sorted by APKI"
for i in sortedByAPKI:
    if args.threadPoolComesFirst:
        pcIndex = highAPKIindices[i] - 1
    else:
        pcIndex = highAPKIindices[i]

    if 0 <= pcIndex and pcIndex < len(pcList):
        print pcList[pcIndex], cumAccesses[highAPKIindices[i]]
    else:
        print "Thread pool", cumAccesses[highAPKIindices[i]]

## Actual clustering

cluster_node.origIndices = highAPKIindices
result =  wpLinkage(shareMissCurves[:, highAPKIindices, :])
print "Max original distance", result[-1,2]
result[:, 2] = result[:,2]/result[-1,2]

### Output clusters ###

for clusters in range(2,9):
    print "Initial target clusters", clusters

    maxClusters = min(clusters, result.shape[0] + 1)
    if result.shape[0] >= maxClusters:
        # take average so that the cut-off is between these two points
        targetDist = (result[-maxClusters, 2] + result[-maxClusters + 1, 2])/2.0
    else:
        # above two constraints imply maxClusters = result.shape[0] + 1 = number of original observations.
        # number of original observations is equal to the number of clusters needed,
        # so just chose a number which is less the distance between the 2 points that were combined the earliest.
        targetDist = result[-maxClusters + 1, 2]*0.9
    print "Target Dist", targetDist, "maxClusters", maxClusters\

    fclusterdata = fcluster(result, targetDist, criterion='distance')
    print "Unoptimized clustering"
    printClusters(fclusterdata, highAPKIindices)

    clusterdata = optimizePools(highAPKIindices, len(pcList), fclusterdata, cumAccesses)

    print "Optimized clustering"
    printClusters(fclusterdata, highAPKIindices)

    out = genPcTableFile(fclusterdata, highAPKIindices, pcList)
    writePcTableFile(out, clusters)

# Write out clustering distances to file
clusterDistances = np.array(result[:,2])
writeClusterDistances(clusterDistances)
